@@ -1,4 +1,3 @@
-
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /* This file is part of the GtkHTML library.
    
@@ -25,6 +24,7 @@
 #include <string.h>
 #include "gtkhtml.h"
 #include "gtkhtml-private.h"
+#include "gtkhtml-stream.h"
 #include "htmlcolorset.h"
 #include "htmlgdkpainter.h"
 #include "htmlprinter.h"
@@ -34,6 +34,8 @@
 #include "htmlselection.h"
 #include "htmlsettings.h"
 #include "htmltokenizer.h"
+#include "htmlembedded.h"
+#include <libgnome/gnome-i18n.h>
 
 #ifndef USE_SCROLLED_WINDOW
 #include <gal/widgets/e-scroll-frame.h>
@@ -42,41 +44,24 @@
 HTMLIFrameClass html_iframe_class;
 static HTMLEmbeddedClass *parent_class = NULL;
 static gboolean calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs);
+	
+static void
+iframe_set_base (GtkHTML *html, const char *url, gpointer data)
+{
+	char *new_url = gtk_html_get_url_base_relative (html, url);
+
+	gtk_html_set_base (html, new_url);
+	g_free (new_url);
+}
 
 static void
 iframe_url_requested (GtkHTML *html, const char *url, GtkHTMLStream *handle, gpointer data)
 {
 	HTMLIFrame *iframe = HTML_IFRAME (data);
 	GtkHTML *parent = GTK_HTML (HTML_EMBEDDED(iframe)->parent);
-	char *new_url = NULL;
 
-	/* FIXME this is not exactly the single safest method of expanding a relative url */
-	if (!strstr (url, ":"))
-		new_url = g_strconcat (iframe->url, url, NULL);
-	
-	gtk_signal_emit_by_name (GTK_OBJECT (parent->engine), "url_requested", new_url ? new_url : url,
-				 handle);
-	
-	if (new_url)
-		g_free (new_url);
-}
-
-static void
-iframe_on_url (GtkHTML *html, const gchar *url, gpointer data)
-{
-	HTMLIFrame *iframe = HTML_IFRAME (data);
-	GtkHTML *parent = GTK_HTML (HTML_EMBEDDED(iframe)->parent);
-
-	gtk_signal_emit_by_name (GTK_OBJECT (parent), "on_url", url);
-}
-
-static void
-iframe_link_clicked (GtkHTML *html, const gchar *url, gpointer data)
-{
-	HTMLIFrame *iframe = HTML_IFRAME (data);
-	GtkHTML *parent = GTK_HTML (HTML_EMBEDDED(iframe)->parent);
-
-	gtk_signal_emit_by_name (GTK_OBJECT (parent), "link_clicked", url);
+	gtk_signal_emit_by_name (GTK_OBJECT (parent->engine), "url_requested",
+				 url, handle);
 }
 
 static void
@@ -145,7 +130,13 @@ static gint
 calc_min_width (HTMLObject *o,
 		HTMLPainter *painter)
 {
-	return html_engine_calc_min_width (GTK_HTML (HTML_IFRAME (o)->html)->engine);
+	HTMLIFrame *iframe;
+
+	iframe = HTML_IFRAME (o);
+	if (iframe->width < 0)
+		return html_engine_calc_min_width (GTK_HTML (HTML_IFRAME (o)->html)->engine);
+	else 
+		return iframe->width;
 }
 
 static void
@@ -245,10 +236,19 @@ copy (HTMLObject *self,
         HTMLIFrame *d = HTML_IFRAME (dest);
 
 	(* HTML_OBJECT_CLASS (parent_class)->copy) (self, dest);
-
+	/*
+	html_iframe_init (s, &html_iframe_class,
+			  HTML_EMBEDDED (dest)->parent,
+			  s->url,
+			  s->width,
+			  s->height,
+			  s->frameborder);
+	*/
 	d->scroll = NULL;
-	d->html = NULL;
+	d->html = gtk_html_new ();
 	d->gdk_painter = NULL;
+	d->old_painter = NULL;
+	d->parent_painter = NULL;
 
 	d->url = g_strdup (s->url);
 	d->width = s->width;
@@ -256,6 +256,24 @@ copy (HTMLObject *self,
 	d->frameborder = s->frameborder;
 }
 
+static HTMLObject *
+op_copy (HTMLObject *self, HTMLObject *parent, HTMLEngine *e, GList *from, GList *to, guint *len)
+{
+	HTMLObject *dup, *clue;
+	GtkHTML *html;
+
+	dup = html_object_dup (self);
+	html = GTK_HTML (HTML_IFRAME (dup)->html);
+	clue = GTK_HTML (HTML_IFRAME (self)->html)->engine->clue;
+	GTK_HTML (HTML_IFRAME (dup)->html)->engine->clue =
+		html_object_op_copy (clue, dup, GTK_HTML (HTML_IFRAME (self)->html)->engine,
+				     html_object_get_bound_list (clue, from),
+				     html_object_get_bound_list (clue, to), len);
+	GTK_HTML (HTML_IFRAME (dup)->html)->engine->clue->parent = parent;
+	
+	return dup;
+}
+
 void
 html_iframe_set_scrolling (HTMLIFrame *iframe, GtkPolicyType scroll)
 {
@@ -290,7 +308,8 @@ calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 		height = html_engine_get_doc_height (e);
 		width = html_engine_get_doc_width (e);
 
-		gtk_widget_set_usize (iframe->scroll, width, height);
+		/* FIXME: fix frames for height > 32767 */
+		gtk_widget_set_usize (iframe->scroll, width, MIN (height, 32767));
 		gtk_widget_queue_resize (iframe->scroll);
 		
 		html_iframe_set_scrolling (iframe, GTK_POLICY_NEVER);
@@ -368,8 +387,8 @@ check_point (HTMLObject *self,
 	    || y >= self->y + self->descent || y < self->y - self->ascent)
 		return NULL;
 
-	x -= self->x + e->leftBorder;
-	y -= self->y - self->ascent + e->topBorder;
+	x -= self->x + e->leftBorder - e->x_offset;
+	y -= self->y - self->ascent + e->topBorder - e->y_offset;
 
 	if (for_cursor && (x < 0 || y < e->clue->y - e->clue->ascent))
 		return html_object_check_point (e->clue, e->painter, 0, e->clue->y - e->clue->ascent,
@@ -395,6 +414,17 @@ append_selection_string (HTMLObject *self,
 	html_object_append_selection_string (GTK_HTML (HTML_IFRAME (self)->html)->engine->clue, buffer);
 }
 
+static void
+reparent (HTMLEmbedded *emb, GtkWidget *html)
+{
+	HTMLIFrame *iframe = HTML_IFRAME (emb);
+
+	gtk_html_set_iframe_parent (GTK_HTML (iframe->html), 
+				    html,
+				    GTK_HTML (iframe->html)->frame);
+	(* HTML_EMBEDDED_CLASS (parent_class)->reparent) (emb, html);
+}
+
 /* static gboolean
 select_range (HTMLObject *self,
 	      HTMLEngine *engine,
@@ -414,7 +444,15 @@ destroy (HTMLObject *o)
 
 	iframe_set_gdk_painter (iframe, NULL);
 
+	g_free (iframe->url);
+
 	if (iframe->html) {
+		if (iframe->old_painter)
+			gtk_object_unref (GTK_OBJECT (iframe->old_painter));
+		if (iframe->parent_painter)
+			gtk_object_unref (GTK_OBJECT (iframe->parent_painter));
+		iframe->old_painter = iframe->parent_painter = NULL;
+
 		gtk_signal_disconnect_by_data (GTK_OBJECT (iframe->html), o);
 		iframe->html = NULL;
 	}
@@ -438,6 +476,7 @@ html_iframe_init (HTMLIFrame *iframe,
 	GtkHTML   *parent_html;
 	GtkHTMLStream *handle;
 	GtkWidget *scrolled_window;
+	gint depth;
 
 	g_assert (GTK_IS_HTML (parent));
 	parent_html = GTK_HTML (parent);
@@ -460,6 +499,12 @@ html_iframe_init (HTMLIFrame *iframe,
 	new_widget = gtk_html_new ();
 	new_html = GTK_HTML (new_widget);
 
+	/* use parent_html font_manager */
+	gtk_object_ref (GTK_OBJECT (new_html->engine->painter));
+	gtk_object_ref (GTK_OBJECT (parent_html->engine->painter));
+	iframe->old_painter = new_html->engine->painter;
+	iframe->parent_painter = parent_html->engine->painter;
+
 	new_tokenizer = html_tokenizer_clone (parent_html->engine->ht);
 
 	html_engine_set_tokenizer (new_html->engine, new_tokenizer);
@@ -469,42 +514,60 @@ html_iframe_init (HTMLIFrame *iframe,
 	gtk_html_set_default_content_type (new_html,
 					   parent_html->priv->content_type);
 	iframe->html = new_widget;
-	gtk_html_set_iframe_parent (new_html, parent, HTML_OBJECT (iframe));
+	depth = gtk_html_set_iframe_parent (new_html, parent, HTML_OBJECT (iframe));
 	gtk_container_add (GTK_CONTAINER (scrolled_window), new_widget);
 	gtk_widget_show (new_widget);
 
-	iframe->url = src;
+	iframe->url = g_strdup (src);
 	iframe->width = width;
 	iframe->height = height;
 	iframe->gdk_painter = NULL;
+	gtk_html_set_base (new_html, src);
 
 	handle = gtk_html_begin (new_html);
+
 	new_html->engine->clue->parent = HTML_OBJECT (iframe);
 
 	gtk_signal_connect (GTK_OBJECT (new_html), "url_requested",
 			    GTK_SIGNAL_FUNC (iframe_url_requested),
 			    (gpointer)iframe);
+#if 0
+	/* NOTE: because of peculiarities of the frame/gtkhtml relationship
+	 * on_url and link_clicked are emitted from the toplevel widget not
+	 * proxied like url_requested is.
+	 */
 	gtk_signal_connect (GTK_OBJECT (new_html), "on_url",
 			    GTK_SIGNAL_FUNC (iframe_on_url), 
 			    (gpointer)iframe);
 	gtk_signal_connect (GTK_OBJECT (new_html), "link_clicked",
 			    GTK_SIGNAL_FUNC (iframe_link_clicked),
 			    (gpointer)iframe);	
+#endif 
 	gtk_signal_connect (GTK_OBJECT (new_html), "size_changed",
 			    GTK_SIGNAL_FUNC (iframe_size_changed),
 			    (gpointer)iframe);	
+	gtk_signal_connect (GTK_OBJECT (new_html), "set_base",
+			    GTK_SIGNAL_FUNC (iframe_set_base),
+			    (gpointer)iframe);	
 	gtk_signal_connect (GTK_OBJECT (new_html), "object_requested",
 			    GTK_SIGNAL_FUNC (iframe_object_requested),
 			    (gpointer)iframe);	
+
 	/*
 	  gtk_signal_connect (GTK_OBJECT (html), "button_press_event",
 	  GTK_SIGNAL_FUNC (iframe_button_press_event), iframe);
 	*/
-	gtk_signal_emit_by_name (GTK_OBJECT (new_html->engine), 
-				 "url_requested", src, handle);
 
-	gtk_widget_set_usize (scrolled_window, width, height);
+	if (depth < 10) {
+		gtk_signal_emit_by_name (GTK_OBJECT (parent_html->engine), 
+					 "url_requested", src, handle);
+	} else {
+		gtk_html_stream_printf (handle, "%s", _("Error: maximium frame depth exceded"));
+		gtk_html_stream_close (handle, GTK_HTML_STREAM_OK);
+	}
 
+	/* FIXME: fix frames for height > 32767 */
+	gtk_widget_set_usize (scrolled_window, width, MIN (height, 32767));
 	gtk_widget_show (scrolled_window);	
 
 	html_embedded_set_widget (em, scrolled_window);
@@ -517,17 +580,14 @@ html_iframe_init (HTMLIFrame *iframe,
 				     parent_html->engine->settings->color_set);
 	html_colorset_set_unchanged (new_html->engine->settings->color_set,
 				     parent_html->engine->settings->color_set);
+	html_painter_set_focus (new_html->engine->painter, parent_html->engine->have_focus);
 	/*
 	gtk_signal_connect (GTK_OBJECT (html), "title_changed",
 			    GTK_SIGNAL_FUNC (title_changed_cb), (gpointer)app);
-	gtk_signal_connect (GTK_OBJECT (html), "set_base",
-			    GTK_SIGNAL_FUNC (on_set_base), (gpointer)app);
 	gtk_signal_connect (GTK_OBJECT (html), "button_press_event",
 			    GTK_SIGNAL_FUNC (on_button_press_event), popup_menu);
 	gtk_signal_connect (GTK_OBJECT (html), "redirect",
 			    GTK_SIGNAL_FUNC (on_redirect), NULL);
-	gtk_signal_connect (GTK_OBJECT (html), "submit",
-			    GTK_SIGNAL_FUNC (on_submit), NULL);
 	gtk_signal_connect (GTK_OBJECT (html), "object_requested",
 			    GTK_SIGNAL_FUNC (object_requested_cmd), NULL);
 	*/
@@ -562,6 +622,7 @@ html_iframe_class_init (HTMLIFrameClass *klass,
 	object_class->reset                   = reset;
 	object_class->draw                    = draw;
 	object_class->copy                    = copy;
+	object_class->op_copy                 = op_copy;
 	object_class->set_max_width           = set_max_width;
 	object_class->forall                  = forall;
 	object_class->check_page_split        = check_page_split;
@@ -573,4 +634,6 @@ html_iframe_class_init (HTMLIFrameClass *klass,
 	object_class->is_container            = is_container;
 	object_class->draw_background         = draw_background;
 	object_class->append_selection_string = append_selection_string;
+	
+	embedded_class->reparent = reparent;
 }
